from pathlib import Path
from typing import Dict, List, Final

import rioxarray
import xarray as xr
import numpy as np

DEFAULT_OUT_OF_SCOPE = 65535

DEFAULT_NO_DATA = 65534

DEFAULT_NO_CROP = 0

# Used for all pixels that are not fields
DEFAULT_NO_CROP_PROBABILITY = 254
DEFAULT_OUTSIDE_AREA_PROBABILITY = 255

DEFAULT_PERENNIAL_CLASSES: Final[List[int]] = [
    2100, 2200, 2310, 2320
]

GRASS_AND_FODDER_CODE: int = 1500
GRASS_AND_FODDER_NAME: str = 'grass/fodder'

DEFAULT_EXCLUDED_VALUES = [
    DEFAULT_NO_CROP, DEFAULT_NO_DATA, DEFAULT_OUT_OF_SCOPE
]

DEFAULT_UNDECIDED_ARABLE_VALUE = 3100
DEFAULT_UNDECIDED_PERENNIAL_VALUE = 3200

DEFAULT_TRANSLATION_TABLE = {
    "1-1-1-0": 1110,  # Wheat
    "1-1-1-1": 1111,  # Winter Wheat
    "1-1-1-2": 1112,  # Spring wheat
    "1-1-2-0": 1120,  # Barley
    "1-1-2-1": 1121,  # Winter Barley
    "1-1-2-2": 1122,  # Spring Barley
    "1-1-3-0": 1130,  # Maize
    "1-1-4-0": 1140,  # Rice
    "1-1-5-0": 1150,  # Other cereals
    "1-1-5-1": 1151,  # Other winter cereals
    "1-1-5-2": 1152,  # Other spring cereals
    "1-2-1-0": 1210,  # Fresh vegetables
    "1-2-2-0": 1220,  # Dry pulses
    "1-3-1-0": 1310,  # Potatoes
    "1-3-2-0": 1320,  # Sugar beet
    "1-4-1-0": 1410,  # Sunflower
    "1-4-2-0": 1420,  # Soybeens
    "1-4-3-0": 1430,  # Rapeseed
    "1-4-4-0": 1440,  # Flax, cotton and hemp
    "1-5-0-0": 1500,  # Grass and fodded crops
    "2-1-0-0": 2100,  # Grapes
    "2-2-0-0": 2200,  # Olives
    "2-3-1-0": 2310,  # Fruits
    "2-3-2-0": 2320,  # Nuts
    "NoCrop": DEFAULT_NO_CROP,
    "NaN": DEFAULT_NO_DATA,
    "nan": DEFAULT_NO_DATA,
    "NAN": DEFAULT_NO_DATA,
    None: DEFAULT_NO_DATA,
}

DEFAULT_HRL_NAMES_TABLE: Dict[int, str] = {
    1110: "wheat",
    1120: "barley",
    1130: "maize",
    1140: "rice",
    1150: "other_cereals",
    1210: "fresh_vegetables",
    1220: "dry_pulses",
    1310: "potatoes",
    1320: "sugar_beet",
    1410: "sunflower",
    1420: "soybeans",
    1430: "rapeseed",
    1440: "flax_cotton_and_hemp",
    1500: "grass/fodder",
    2100: "grapes",
    2200: "olives",
    2310: "fruits",
    2320: "nuts"
}

DEFAULT_HRL_NAMES_FULL_TABLE: Dict[int, str] = {
    1110: "Wheat",
    1120: "Barley",
    1130: "Maize",
    1140: "Rice",
    1150: "Other cereals (rye, oats, triticale)",
    1210: "Fresh Vegetables",
    1220: "Dry Pulses",
    1310: "Potatoes",
    1320: "Sugar Beet",
    1410: "Sunflower",
    1420: "Soybeans",
    1430: "Rapeseed",
    1440: "Flax, cotton and hemp",
    1500: "Grass/fodder",
    2100: "Grapes",
    2200: "Olives",
    2310: "Fruits",
    2320: "Nuts",
    # Undecided and other special classes
    DEFAULT_UNDECIDED_ARABLE_VALUE: "undecided_arable",
    DEFAULT_UNDECIDED_PERENNIAL_VALUE: "undecided_perennial",
}

DEFAULT_REGROUPMENT_TABLE: Dict[int, int] = {
    111: 1110,
    112: 1120,
    113: 1130,
    114: 1140,
    115: 1150,
    121: 1210,
    122: 1220,
    131: 1310,
    132: 1320,
    141: 1410,
    142: 1420,
    143: 1430,
    144: 1440,
    150: 1500,
    210: 2100,
    220: 2200,
    231: 2310,
    232: 2320,
}


DEFAULT_CLASS_NAMES = list(DEFAULT_HRL_NAMES_TABLE.values())


class InvalidValuesError(Exception):
    def __init__(self, values_counts) -> None:
        self._values_counts = values_counts
        super().__init__(f'Invalid values detected in postprocessed raster: {self._values_counts}')

    def __str__(self):
        return '\n'.join([f'Values: {val} | Count: {count}' for (val, count) in self._values_counts])


def translate_layer(inarr: xr.DataArray, translation_table: Dict[str, int] = None) -> xr.DataArray:
    """
    Translates the crop type values predicted by the model (string type) into HRL code using the given
    translation table. Probabilities are not changed.

    :param inarr: The input DataArray to process, needs at least one band named `croptype`
    :param translation_table: A mapping string -> int to perform changes in the croptype labels.
    """
    translation_table = DEFAULT_TRANSLATION_TABLE if translation_table is None else translation_table

    def map_func(input_data: np.ndarray, translation_table: Dict[str, int]):
        output_data = np.zeros_like(input_data, dtype=np.uint16)
        for orig, dst in translation_table.items():
            output_data[input_data == orig] = dst
        return output_data

    # Two modes, either the input is a single band array, in that case no selection is required, or the input array is
    # multiband. In that case, selection of the first band is required.

    if 'bands' not in inarr.dims:
        translated_data = xr.apply_ufunc(map_func, inarr, kwargs={
                                         'translation_table': translation_table})
    else:
        translated_layer = xr.apply_ufunc(
            map_func,
            inarr.sel({'bands': 'croptype'}),
            kwargs={'translation_table': translation_table},
            dask='parallelized'
        )
        translated_data = inarr.copy()
        translated_data.loc[{'bands': 'croptype'}] = translated_layer

    return translated_data


def _setup_aggregation_template(
    inarr: xr.DataArray,
    regroupment_table: Dict[int, int] = DEFAULT_REGROUPMENT_TABLE,
    hrl_names: Dict[int, str] = DEFAULT_HRL_NAMES_TABLE
) -> xr.DataArray:
    """
    Returns a Xarray with empty elements but with shape/coordinates that match
    the output of an aggregation. This is necessary for Dask to execute the 
    aggregation method, as dimensions and coordinates of the output must be
    known before execution to be able to do lazy computation.
    """

    n_groups = len(regroupment_table.values())
    empty_array = np.empty((2 + n_groups, *inarr.shape[1:]), dtype=np.uint16)

    band_names = ['croptype', 'probability']
    for group_nb in regroupment_table.values():
        croptype_name = hrl_names[group_nb]
        band_names.append(f'probability_{croptype_name}')

    return xr.DataArray(
        data=empty_array,
        dims=('bands', 'y', 'x'),
        coords={
            'bands': band_names,
            'y': inarr.coords['y'],
            'x': inarr.coords['x']
        }
    )

def aggregate_croptypes(
    inarr: xr.DataArray,
    regroupment_table: Dict[int, int] = DEFAULT_REGROUPMENT_TABLE,
    hrl_names: Dict[int, str] = DEFAULT_HRL_NAMES_TABLE,
) -> xr.DataArray:
    """
    Using the regroupment table, croptypes will be aggregated in groups.

    :param inarr: The input DataArray to process, needs at least one band named
                  `croptype` and one band named `max_probability`. If there is
                  a probability band per crop type, then the probability bands
                  must be named `probability_<class_name>`, where class names 
                  are lowercased, spaces are remplaced by underscores (_), and
                  commas (,) are removed.
    :param regroupment_table: A table regroupping crop types. Keys are tuples of
                              crop types in and values are target group codes.
                              Both keys/values contains HRL codes in integer.
    :param hrl_names: A table mapping HRL codes in integer with their class 
                      names.
    """
    # Getting the unique values of crop types present in the raster
    present_croptypes = np.unique(
        inarr.sel(bands='croptype').to_numpy()
    )

    # Creation of new probability array
    newprob_shape = (len(regroupment_table.values()), *inarr.shape[1:])
    new_probabilities = np.zeros(newprob_shape, dtype=np.uint8)
    new_croptype = np.zeros(inarr.shape[1:], dtype=np.uint16)

    # Track the pixels that where selected for modification
    modified_pixels = np.zeros(inarr.shape[1:], dtype=np.bool_)

    # xarray DataArray containing the crop types
    croptype = inarr.sel(bands='croptype').to_numpy()

    for group_idx, (croptype_suffix, croptype_group) in enumerate(regroupment_table.items()):
        # Finding the croptypes associated from the suffix
        croptype_classes = [
            cropclass for cropclass in present_croptypes
            if str(cropclass).startswith(str(croptype_suffix))
        ]

        # Masking selecting pixels with croptype of the current group of classes
        croptype_mask = np.isin(croptype, croptype_classes)
        modified_pixels = np.bitwise_or(modified_pixels, croptype_mask)
        new_croptype[croptype_mask] = croptype_group

        # Selecting the probabilities of the group
        probability_band_names = [
            f'probability_{hrl_names[cropclass]}'
            for cropclass in croptype_classes
        ]

        # Checking if those probabilities are available in the raster
        probability_band_names = list(
            set(probability_band_names).intersection(set(
                inarr.coords['bands'].to_numpy()))
        )

        probabilities_for_group = inarr.sel(
            bands=probability_band_names
        ).sum(dim='bands')

        new_probabilities[group_idx] = probabilities_for_group

    # Resets untouched pixels to the new croptype array
    new_croptype[~modified_pixels] = croptype[~modified_pixels]

    # Computation of max probablities
    new_probability = np.max(new_probabilities, axis=0)

    # Get group names
    all_group_names = [
        f'probability_{hrl_names[group]}'
        for group in regroupment_table.values()
    ]

    # Creation of an array with new probabilities
    outarr = xr.DataArray(
        data=np.concatenate([
            new_croptype[np.newaxis],
            new_probability[np.newaxis],
            new_probabilities
        ]),
        dims=inarr.dims,
        coords={
            'x': inarr.coords['x'],
            'y': inarr.coords['y'],
            'bands': ['croptype', 'probability', *all_group_names]
        }
    )

    return outarr


def apply_class_names(
    inarr: xr.DataArray,
    class_names: List[str] = DEFAULT_CLASS_NAMES,
) -> xr.DataArray:
    """
    Reapplies the name on the croptype, max probability and per-class
    probability layers. Useful as the DataArrays read from a GeoTiff file loose
    their band names. Band names are necessary for postprocessing tasks,
    especially for probabilities.

    Parameters
    ----------
    inarr: xr.DataArray
        The array without names on the bands coordinate to apply the names on.
    class_names: Dict[str]
        The names of the classes, the size of the lsit must match the number of
        bands of the input array minus 2 (for croptype and max probability
        layer)
    Returns
    -------
    outarr: xr.DataArray
        An array with the names applied on  
    """
    all_band_names = ['croptype', 'max_probability']
    for cls_name in class_names:
        all_band_names.append(f'probability_{cls_name}')
    return inarr.assign_coords({'bands': all_band_names})

def remask_invalid_croptypes(inarr: xr.DataArray) -> xr.DataArray:
    """ Changes the value of 10'000 for the croptype layer to no crop. 
    """
    outarr = inarr.copy()

    croptype = outarr.sel(bands='croptype')

    nocrop_mask = (croptype == 10000) | (croptype == DEFAULT_NO_CROP)

    outarr.sel(bands='croptype')[:] = outarr.sel(bands='croptype').where(
        ~nocrop_mask, DEFAULT_NO_CROP
    )
    outarr.isel(bands=slice(1, None))[:] = outarr.isel(bands=slice(1, None)).where(
        ~nocrop_mask, DEFAULT_NO_CROP_PROBABILITY
    )

    return outarr

def apply_bvl(
    inarr: xr.DataArray,
    bvl_mask_path: Path,
    perennial_classes: List[int] = DEFAULT_PERENNIAL_CLASSES
) -> xr.DataArray:
    """ Uses the Base Vegetation Layer (BVL) product to exclude areas of no-crop
    or no-data from the crop type product. Additionaly, the BVL layer can
    improve perennial fields classified as grass and fodder by croptype.

    Parameters
    ----------
    inarr: xr.DataArray
        Input array to apply the BVL mask to.
    bvl_mask_path: Path
        Path to the BVL layer to load and apply.
    perennial_classes: List[int]
        HRL Product codes for the perennial classes.
    Returns
    -------
    outarr: xr.DataArray
        Masked array by the BVL. 
    """

    # Loading and aligning the bvl mask to the
    bvl_mask = rioxarray.open_rasterio(
        bvl_mask_path
    ).astype('ubyte').isel(band=0)

    # Renaming dimension band to bands to match new repo name standards
    if 'band' in inarr.dims:
        inarr = inarr.rename({'band': 'bands'})

    # Extracting the bvl mask section intersecting with our array (same CRS)
    bvl_mask = bvl_mask.rio.clip_box(*inarr.rio.bounds())

    #for DOM area's sometimes a padding needs to be added.
    bvl_mask = bvl_mask.rio.pad_box(*inarr.rio.bounds())

    assert (
        (bvl_mask.coords['x'] == inarr.coords['x']).all()
        and (bvl_mask.coords['y'] == inarr.coords['y']).all()
    )

    # Creates a binary mask from the BVL layer
    nocropland_mask = (bvl_mask < 4) | (bvl_mask == 6) | (bvl_mask == 8)
    outside_area_mask = bvl_mask == 255

    # Apply masking on croptype
    inarr.isel(bands=0)[:] = xr.where(
        nocropland_mask, DEFAULT_NO_CROP, inarr.isel(bands=0)
    )
    inarr.isel(bands=0)[:] = xr.where(
        outside_area_mask, DEFAULT_OUT_OF_SCOPE, inarr.isel(bands=0)
    )

    # Apply masking on probabilities
    inarr.isel(bands=slice(1, None))[:] = xr.where(
        nocropland_mask, DEFAULT_NO_CROP_PROBABILITY, inarr.isel(bands=slice(1, None))
    ).transpose('bands', 'y', 'x')
    inarr.isel(bands=slice(1, None))[:] = xr.where(
        outside_area_mask, DEFAULT_OUTSIDE_AREA_PROBABILITY, inarr.isel(bands=slice(1, None))
    ).transpose('bands', 'y', 'x')

    # Fix zones classified as tree crops in BVL but classified as grass and
    # fodder in the croptype layer
    grass_fodder_mask = inarr.isel(bands=0) == GRASS_AND_FODDER_CODE
    misclassified_grass_fodder = grass_fodder_mask & (bvl_mask == 5)

    # For every pixel of the tree classes, see which has the highest probability
    perennial_class_names = [
        DEFAULT_HRL_NAMES_TABLE[perennial_id]
        for perennial_id in perennial_classes
    ]
    probability_perennial_band_names = [
        f'probability_{perennial_name}'
        for perennial_name in perennial_class_names
    ]

    # Changes the croptype of the misdetected grass to the perennial class
    # with the highest probability, then swap the grass/fodder class probability
    # to the winning perennial class.
    perennial_probabilities = inarr.sel(
        bands=probability_perennial_band_names
    )
    grass_fodder_probability = inarr.sel(
        bands=f'probability_{GRASS_AND_FODDER_NAME}'
    )

    highest_perennial_index = perennial_probabilities.argmax(dim='bands')

    new_perennial_croptype = \
        np.array(perennial_classes)[highest_perennial_index]
    best_perennial_probability = \
        perennial_probabilities[highest_perennial_index, :, :]

    # Change croptype to winning class
    inarr.isel(bands=0)[:] = inarr.isel(bands=0).where(
        ~misclassified_grass_fodder, new_perennial_croptype
    )

    # Swap probability of grass/fodder with probability of winning perennial
    # Keep the grass and fodder probability
    grass_fodder_probability_backup = grass_fodder_probability.copy()

    grass_fodder_probability[:] = grass_fodder_probability.where(
        ~misclassified_grass_fodder, best_perennial_probability
    )
    best_perennial_probability[:] = best_perennial_probability.where(
        ~misclassified_grass_fodder, grass_fodder_probability_backup
    )

    return inarr

def remove_grass_and_fodder(
    inarr: xr.DataArray,
    class_id: int = GRASS_AND_FODDER_CODE
) -> xr.DataArray:
    """ Sets the croptype and the probability of the pixels classified as grass
        and fodder to No Cropland. Only croptype and max_probability will be
        changed, not the per class probabilities (if included in the provided
        input array)

        Parameters
        ----------
        inarr: xr.DataArray
            A DataArray to modify with the croptype as the first band and the
            max_probability as the second
        class_id: int
            The integer value of the grass and fodder class
        Returns
        -------
        outarr: xr.DataArray
            Input array with the grass and fodder class masked
    """
    grass_fodder_mask = inarr.isel(bands=0) == class_id

    inarr.isel(bands=0)[:] = xr.where(
        grass_fodder_mask, DEFAULT_NO_CROP, inarr.isel(bands=0)
    )
    inarr.isel(bands=slice(1, None))[:] = xr.where(
        grass_fodder_mask, DEFAULT_NO_CROP_PROBABILITY,
        inarr.isel(bands=slice(1, None))
    ).transpose('bands', 'y', 'x')

    return inarr

def postprocessed_tile_check(
    input: xr.DataArray,
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES,
    excluded_values_probabilities: List[int] = [DEFAULT_NO_CROP_PROBABILITY, DEFAULT_OUTSIDE_AREA_PROBABILITY],
    class_ids: List[int] = list(DEFAULT_HRL_NAMES_FULL_TABLE.keys())
) -> None:
    """ Performs checks on the croptype, seeing if all the values are either
        croptype or known excluded values. Also performs checks on the
        probability, seeing if there are included between [0, 100] and or in
        the probability excluded values. 
    """
    croptype = input.sel(bands='croptype')
    probability = input.sel(bands='max_probability')
    
    # Checks the values for the croptype layer
    accepted_values = class_ids + excluded_values
    unvalid_croptype_mask = ~(croptype.isin(accepted_values))
    
    if (unvalid_croptype_mask * 1).sum() > 0:
        # Compute statistics
        unvalid_croptype_values = croptype.where(
            unvalid_croptype_mask, drop=True
        )
        values, counts = np.unique(unvalid_croptype_values, return_counts=True)
        raise InvalidValuesError(np.transpose(np.stack([values, counts])))

    accepted_probability_values = \
        list(np.arange(101)) + excluded_values_probabilities

    unvalid_probability_mask = ~(probability.isin(accepted_probability_values))
    
    if (unvalid_probability_mask * 1).sum() > 0:
        # Compute statistics
        unvalid_probability_values = probability.where(
            unvalid_probability_mask, drop=True
        )
        values, counts = np.unique(
            unvalid_probability_values, return_counts=True
        )
        raise InvalidValuesError(np.transpose(np.stack([values, counts])))

def compute_tile_statistics(
    input: xr.DataArray,
    class_ids: List[int]= list(DEFAULT_HRL_NAMES_FULL_TABLE.keys())
) -> dict:
    """ Computes statistics such as the percentage of crop area and percentage
        of undecided pixels over that crop area
    """
    croptype = input.sel(bands='croptype')
    
    crop_pixels = np.isin(croptype, class_ids) * 1

    crop_percentage = crop_pixels.sum() / (croptype.data.shape[0] * croptype.data.shape[1])

    undecided_pixels = np.isin(
        croptype,
        [DEFAULT_UNDECIDED_ARABLE_VALUE, DEFAULT_UNDECIDED_PERENNIAL_VALUE]
    )

    undecided_percentage = undecided_pixels.sum() / crop_pixels.sum()

    return {
        'crop_land_ratio': crop_percentage,
        'undecided_crop_ratio': undecided_percentage
    }
