import time
from typing import List, Literal, Optional
import tempfile
import subprocess

import rasterio
import numpy as np

from loguru import logger
from tqdm import tqdm

from cropclass.postprocessing.layer_format import (
    DEFAULT_EXCLUDED_VALUES,
    DEFAULT_UNDECIDED_ARABLE_VALUE,
    DEFAULT_UNDECIDED_PERENNIAL_VALUE,
    DEFAULT_NO_CROP,
    DEFAULT_NO_CROP_PROBABILITY,
    DEFAULT_OUTSIDE_AREA_PROBABILITY,
    DEFAULT_NO_DATA,
    DEFAULT_OUT_OF_SCOPE,
    DEFAULT_HRL_NAMES_TABLE,
    GRASS_AND_FODDER_CODE
)

from cropclass.postprocessing.layer_format import (
    DEFAULT_HRL_NAMES_TABLE
)

CHUNK_X_SIZE, CHUNK_Y_SIZE = 128, 128
CHUNK_X_OVERLAP, CHUNK_Y_OVERLAP = 16, 16

class InvalidProbabilityError(Exception):
    pass

def _check_probability(
    input_data: np.ndarray,
    phase_name: str,
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES
) -> np.ndarray:
    """ Checks the coherence of prediction data in three ways.
        1. Checks if no max probability is above 100
        2. If there are per-class probabilities, checks that the sum of those
           per class probability is 100.
        3. If there are per-class probabilities, checks if the highest of those
           probabilities is reflected by max probability.

        It is asumed at the parameter input_data is a 3D array where the first
        dimension/axis has minimum two elements (croptype and max probability),
        if it has more than two elements, then the per-class probabilties are
        supposed to be the rest of the elements
    """
    if input_data.shape[0] < 2:
        return input_data
    # Invalid mask (mask for pixels that shouldn't be checked)
    invalid_mask = np.isin(input_data[0], excluded_values)

    # Checks if the max_probability is not over 100
    probabilties_over_100 = input_data[1] > 100
    if (probabilties_over_100 & ~invalid_mask * 1).sum() > 0:
        raise InvalidProbabilityError(
            f'Phase: {phase_name} - Some pixels have max_probability over 100 ' 
            f'(max: {np.max(input_data[1])})'
        )
    if input_data.shape[0] > 2:
        # Assumes that the arrays starting from the third one are per-class 
        # probability. Checks if the sum of those probabilities is not over 100
        sum_class_probabilities = np.sum(input_data[2:], axis=0)
        sum_class_probabilities_over_100 = sum_class_probabilities > 105  # Rounding error tolerance
        if (sum_class_probabilities_over_100 & ~invalid_mask * 1).sum() > 0:
            raise InvalidProbabilityError(
                f'Phase {phase_name} - Sum of per class probabilities lead to'
                f' higher probabilities than 100 '
                f'({np.max(sum_class_probabilities)})'
            )

        # Checks if max probability reflects the higher value in the per class
        # probabilities
        highest_class_probability = np.max(input_data[2:], axis=0)

        non_matching_pixels = abs(
            input_data[1].astype(int) - highest_class_probability.astype(int)
        ) > 1
        non_match_pixels_count = (non_matching_pixels & ~invalid_mask * 1).sum()
        if non_match_pixels_count > 0:
            raise InvalidProbabilityError(
                f'Phase: {phase_name} - Max probability doesn\'t match the '
                f'highest probability in the classes probability for some '
                f'pixels ({non_match_pixels_count})'
            )
    return input_data


def smooth_and_reclassify(
    input_data: np.ndarray,
    aggregated_class_codes: Optional[List[int]] = None,
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES
) -> np.ndarray:
    """
    This function smooths all the prediction scores of a croptype tile to produce
    a new croptype tile with new predicted values.
    
    It is assumed that the probabilities of all classes are included.

    :param input_data: 3D array with the 'band' dimension in the first axis. 
                       The first axis must contain the prediciton, max_probability
                       and a probability for every crop type class
    :param aggreagated_class_names: The names of the hrl classes, must be one
                                    per class probability in the input array and
                                    be in the right order.
    :param excluded_values: A list of values that are either NoCrop, NoData or OutOfScope
    """
    from scipy.signal import convolve2d
    from cropclass.postprocessing.layer_format import DEFAULT_REGROUPMENT_TABLE


    if aggregated_class_codes is None:
        aggregated_class_codes = DEFAULT_REGROUPMENT_TABLE.values()

    index_to_croptype = dict(enumerate(aggregated_class_codes))

    prediction = input_data[0]
    max_probability = input_data[1]

    # All probabilities are assumed to exist here and be in the same number
    # as the given translation table
    all_probabilities = input_data[2:]

    assert all_probabilities.shape[0] == len(aggregated_class_codes),\
         f'The number of class probabilities {all_probabilities.shape[0]} is ' \
         f'not equals to the number of crop types in the translation table ' \
         f'{len(index_to_croptype)}'

    # computes the excluded mask of values that remain unchanged
    excluded_mask = np.in1d(
        prediction.reshape(-1),
        excluded_values
    ).reshape(*prediction.shape)

    # Sets up the elements of all_probabilities array to 0 with the excluded_mask
    all_probabilities[:, excluded_mask] = 0

    conv_kernel = np.array([
        [1, 2, 1],
        [2, 3, 2],
        [1, 2, 1]
    ], dtype=np.int16)

    # Performs smoothing for all classes
    for cls_idx, _ in index_to_croptype.items():
        all_probabilities[cls_idx] = convolve2d(
            all_probabilities[cls_idx],
            conv_kernel,
            mode='same',
            boundary='symm'
        ) / conv_kernel.sum()
        all_probabilities[cls_idx] = (all_probabilities[cls_idx] + 0.5).astype(np.uint16)
        all_probabilities[cls_idx][excluded_mask] = 0

    # Compute new max probability
    new_max_probability = np.amax(all_probabilities, axis=0)
    new_max_probability[excluded_mask] = max_probability[excluded_mask]

    # Compute the new crop type
    new_prediction_idx = np.argmax(all_probabilities, axis=0)
    new_prediction = prediction.copy()
    for cls_idx, hrl_code in index_to_croptype.items():
        current_index_mask = new_prediction_idx == cls_idx
        new_prediction[current_index_mask] = hrl_code

    new_prediction[excluded_mask] = prediction[excluded_mask]

    return np.array([new_prediction, new_max_probability, *all_probabilities])


def remove_small_objects(input_data: np.ndarray,
                         target_excluded_value: int = DEFAULT_NO_CROP,
                         target_excluded_value_probs: int = DEFAULT_NO_CROP_PROBABILITY,
                         minimum_object_size: int = 25,
                         excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES,
                         excluded_values_probs: List[int] = [DEFAULT_NO_CROP_PROBABILITY, DEFAULT_OUTSIDE_AREA_PROBABILITY]) -> np.ndarray:
    """
    Excludes objects (blobs of pixels) that have an area below 25 pixels. Here, an object is considered to be a
    classified crop field, or an island of crop fields that is surrounded by no-crop land. The excluded pixels
    are set to the specified target_exclude_value, and the probabilities are set to 0.

    :param input_data: A numpy array containing in the first dimensions two elements: the class codes and the associated
                       predicted probability/confidence.
    :param target_excluded_value: Value on which to convert isolated pixels.
    :param minimum_object_size: The minimum size of the crop blobs to be kept
    :param excluded_values: Values that are considered as non-crop land cover or invalid data
    :return: An array containing the processed results as well as an array with the new prediction probabilities.
    """

    from skimage.morphology import remove_small_objects

    prediction = input_data[0]
    probabilities = input_data[1:]

    # Creates a binary mask containing the excluded values
    excluded_values_mask = np.zeros(shape=prediction.shape, dtype=np.bool)
    for excluded_val in excluded_values:
        excluded_values_mask[prediction == excluded_val] = True
    valid_mask = ~excluded_values_mask

    small_fields_mask = np.zeros(shape=input_data.shape[1:], dtype=bool)

    # Iterates for each class
    all_classes = np.unique(prediction)
    # For every class, see which fields are too small
    for class_id in all_classes:
        class_related_fields = (prediction == class_id) & valid_mask
        class_related_smallfields_removed = remove_small_objects(
            class_related_fields, minimum_object_size
        )
        # Small class fields are the XOR between all class fields and the
        # class fields with small fields removed
        class_small_fields_mask = np.bitwise_xor(
            class_related_fields, class_related_smallfields_removed
        )
        assert class_small_fields_mask.sum() == class_related_fields.sum() - \
            class_related_smallfields_removed.sum()
        # Append this to the small objects mask
        small_fields_mask = small_fields_mask | class_small_fields_mask

    valid_mask[small_fields_mask] = 0

    prediction_cleaned = prediction.copy()
    probabilities_cleaned = probabilities.copy()

    prediction_cleaned[~valid_mask] = target_excluded_value
    probabilities_cleaned[:, ~valid_mask] = target_excluded_value_probs

    # Makes sure that previously excluded values remains excluded
    for excluded_val in excluded_values:
        prediction_cleaned[prediction == excluded_val] = excluded_val
    for excluded_val in excluded_values_probs:
        probabilities_cleaned[0, probabilities[0] == excluded_val] = excluded_val

    return np.array([prediction_cleaned, *probabilities_cleaned])


def sieve_reimplementation(
    input_data: np.ndarray,
    minimum_object_size: int = DEFAULT_NO_CROP,
    connectivity: Literal[4, 8] = 8,
    excluded_values: List[int] = [DEFAULT_NO_DATA, DEFAULT_OUT_OF_SCOPE]
) -> np.ndarray:
    """
    Is a combination between remove_isolated_pixels and majority_voting,
    changing fields blobs that are too small into the neighbor majority class.

    Assigned max probability will be the mean probability of the selected
    neighbors, but the process loosses consistency in per-class probability.

    :param input_data: The input array containing in the first axis the, 
    prediction, max_probability and per-class probabilities 
    :param min_object_size: Area expressed in pixels under whih the fields will
    be subject of change.
    :param excluded_values: Excluded values, pixels under those values will not
    be considered as fields.
    """
    from skimage.measure import regionprops, label
    from skimage.morphology import binary_dilation
    import skimage.morphology

    predictions = input_data[0]
    probabilities = input_data[1:]

    # Configures the kernel
    if connectivity == 4:
        neighbors_kernel = np.array(
            [[0, 1, 0], [1, 0, 1], [0, 1, 0]], dtype='uint8'
        )
    elif connectivity == 8:
        neighbors_kernel = np.ones((3, 3), dtype='uint8')
        neighbors_kernel[1, 1] = 0
    else:
        raise ValueError(
            'Argument connectivity can either be set to 4 or 8, '
            f'here is set to: {connectivity}'
        )

    # Save the excluded values as they will be kept excluded.
    target_excluded_mask = np.isin(predictions, excluded_values)

    output_prediciton = predictions.copy()
    output_probabilities = probabilities.copy()

    # Compute mask of unique objects
    fields = label(predictions)
    # Remove excluded values
    fields[target_excluded_mask] = 0

    # Remove small objects in the fields
    large_fields = skimage.morphology.remove_small_objects(
        fields, min_size=minimum_object_size, connectivity=1
    )

    # Get the small field mask, the one to modify
    small_fields_mask = fields != large_fields

    fields[~small_fields_mask] = 0

    # Iterates over the small fields and perform neighbor voting
    for field_id in tqdm(np.unique(fields)):
        if field_id == 0:  # Background class
            continue
        
        # Get binary array representing the pixels that are the neighbors from
        # the field.
        field_mask = fields == field_id
        assert field_mask.sum() < minimum_object_size  # Make sure large fields are not selected
        neighbors = np.bitwise_xor(
            binary_dilation(field_mask, footprint=neighbors_kernel),
            field_mask
        )

        neighbor_values = predictions[neighbors]
        neighbor_probs = probabilities[0][neighbors]

        # Finding which is the winning neighbor
        croptype_values, counts = np.unique(neighbor_values, return_counts=True)
        winning_neighbor = croptype_values[np.argmax(counts)]

        # Changing the field values and the mean probability
        output_prediciton[field_mask] = winning_neighbor
        output_probabilities[0][field_mask] = (np.mean(neighbor_probs) + 0.5).astype('uint16')


    # All the field classes (not )
    output_prediciton[target_excluded_mask] = predictions[target_excluded_mask]
    output_probabilities[:, target_excluded_mask] = \
        probabilities[:, target_excluded_mask]

    return np.array([output_prediciton, *output_probabilities])


def sieve(
    input_data: np.ndarray,
    minimum_object_size: int = 25,
    connectivity: Literal[4, 8] = 8,
    excluded_values: List[int] = [DEFAULT_NO_DATA, DEFAULT_OUT_OF_SCOPE],  # No crop is still kept
    class_ids: List[int] = list(DEFAULT_HRL_NAMES_TABLE.keys())
) -> np.ndarray:
    """ Calls directly GDAL sieve from a subprocess, writing the input data
        in a temporary file.
    """

    croptype = input_data[0]
    max_probability = input_data[1]
    per_class_probabilities = input_data[2:]

    if connectivity not in [4, 8]:
        raise ValueError('Connectivity parameter can be either 4 or 8.')

    # Checks is the amount of per class probability per pixels matches the size
    # of the given class ids.
    if len(per_class_probabilities) != len(class_ids):
        raise ValueError(
            'Given class ids does not correspond to the amount of per class '
            'probability layers.'
        )

    valid_mask = ~(np.isin(croptype, excluded_values))

    input_file = tempfile.NamedTemporaryFile()
    mask_file = tempfile.NamedTemporaryFile()
    output_file = tempfile.NamedTemporaryFile()

    # Write input data to disk as well as mask
    with rasterio.open(
        input_file, 'w', driver='GTiff', width=croptype.shape[1],
        height=croptype.shape[0], count=1, dtype='uint16'
    ) as dst:
        dst.write(croptype, 1)

    with rasterio.open(
        mask_file, 'w', driver='GTiff', width=valid_mask.shape[1],
        height=valid_mask.shape[0], count=1, dtype='uint16'
    ) as msk:
        msk.write(valid_mask.astype('uint8'), 1)

    # Call command
    gdal_command = f'gdal_sieve.py -st {minimum_object_size} -{connectivity} ' \
                   f'{input_file.name} -mask {mask_file.name} {output_file.name}'

    result = subprocess.run(
        gdal_command, shell=True, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, text=True
    )

    if result.returncode != 0:
        logger.error('Problem encountered when using GDAL_sieve, printing logs')
        logger.error('STOUD output -------------------------------------')
        logger.error(result.stdout)
        logger.error('STDERR output ------------------------------------')
        logger.error(result.stderr)
        logger.error('END of logs --------------------------------------')
        raise Exception('Problem encountered when using GDAL_sieve.')

    # Loading the output data
    with rasterio.open(output_file, 'r', driver='GTIFF') as src:
        processed_croptype = src.read(1)
    
    # Compute the difference array, showing which pixels where modified and
    # require a change in probabilities. Also ignores the perennial and arable
    # values.
    change_array = (processed_croptype != croptype) & \
        (processed_croptype != DEFAULT_UNDECIDED_PERENNIAL_VALUE) & \
        (processed_croptype != DEFAULT_UNDECIDED_ARABLE_VALUE)
    
    new_croptype_values = processed_croptype[change_array]

    # Get indices
    probability_indices = np.searchsorted(
        class_ids, new_croptype_values
    )

    # For each changed pixel, changes the max probability of the pixel from the
    # class probability of the new croptype.
    output_probability = max_probability.copy()
    output_probability[change_array] = \
        per_class_probabilities[:, change_array][probability_indices, np.arange(len(probability_indices))]
    
    # Mask the probabilities of the pixels that are excluded
    output_probability[processed_croptype == DEFAULT_NO_CROP] = \
        DEFAULT_NO_CROP_PROBABILITY
    output_probability[processed_croptype == DEFAULT_OUT_OF_SCOPE] = \
        DEFAULT_OUTSIDE_AREA_PROBABILITY

    input_file.close()
    mask_file.close()
    output_file.close()

    return np.array([processed_croptype, output_probability, *per_class_probabilities])


def remove_isolated_pixels(input_data: np.ndarray,
                           target_excluded_value: int = DEFAULT_NO_CROP,
                           number_of_neighbors: int = 5,
                           excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES) -> np.ndarray:
    """
    Excluding the pixels classified as crops that have less neighbors than the specified parameter. Their probability
    are set to 0.

    :param input_data: A numpy array containing in the first dimensions two elements: the class codes and the associated
                       predicted probability/confidence.
    :param target_excluded_value: Value on which to convert isolated pixels.
    :param number_of_neighbors: Number of minimum neighbors to have to remain valid.
    :param excluded_values: Values that are considered as non-crop land cover or invalid data
    :return: An array containing the processed results as well as an array with the new prediction probabilities.
    """
    from scipy.signal import convolve2d

    prediction = input_data[0]
    probability = input_data[1]

    # Creates mask containing all the excluded values
    exclusion_mask = np.zeros(shape=prediction.shape, dtype=np.bool)
    for excluded_val in excluded_values:
        exclusion_mask[prediction == excluded_val] = True

    # Inverts the mask, creating a mask that includes valid pixels
    valid_mask = (~exclusion_mask).astype(np.uint8)

    # 3x3 kernel, used for counting neighbor valid pixels (8 neighbors around each pixel)
    kernel = np.ones(shape=(3, 3), dtype=np.uint8)
    # Setting the central value of the kernel to 0 (we don't count the pixel itself as a neighbor)
    kernel[1, 1] = 0

    # Performs convolution, essentially counting the neighbors for each pixel
    neighbor_counts = convolve2d(valid_mask, kernel, mode='same', fillvalue=0)  # TODO: evaluate impact of fillvalue

    # Create the final isolated mask and return a copy of the input data but with isolated pixels marked as excluded
    isolated_pixels = neighbor_counts < number_of_neighbors
    output = prediction.copy()
    output_proba = probability.copy()

    output[isolated_pixels] = target_excluded_value
    output_proba[isolated_pixels] = 0

    # The pixels that were originally considered as one of the excluded classes are reverted back to this original class
    # (For example, we don't want to change nodata pixels into nocrop pixels)
    for excluded_val in excluded_values:
        output[prediction == excluded_val] = excluded_val
        output_proba[prediction == excluded_val] = 0

    return np.array([output, output_proba, *input_data[2:]])


def majority_vote(input_data: np.ndarray, kernel_size: int = 7, conf_threshold: int = 30,
                  target_excluded_value: int = DEFAULT_NO_CROP,
                  excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES) -> np.ndarray:
    """
    Majority vote is performed using a sliding local kernel. For each pixel, the voting of a final class is done from
    neighbours values weighted with the confidence threshold. Pixels that have one of the specified excluded values are
    excluded in the voting process and are unchanged.

    The prediction probabilities are reevaluated by taking, for each pixel, the average of probabilities of the
    neighbors that belong to the winning class. (For example, if a pixel was voted to class 2 and there are three
    neighbors of that class, then the new probability is the sum of the old probabilities of each pixels divided by 3)

    :param input_data: A numpy array containing in the first dimensions two elements: the class codes and the associated
                       predicted probability/confidence.
    :param kernel_size: The size of the kernel used for the neighbour around the pixel.
    :param conf_threshold: Pixels under this confidence threshold do not count into the voting process.
    :param target_excluded_value: Pixels that have a null score for every class are turned into this exclusion value
    :param excluded_values: Pixels that have on of the excluded values do not count into the voting process and are
                            unchanged.
    :return: An array containing the processed results as well as an array with the new prediction probabilities.
    """
    from scipy.signal import convolve2d

    # As the probabilities are in integers between 0 and 100, we use uint16 matrices to store the vote scores
    assert kernel_size <= 25, f'Kernel value cannot be larger than 25 (currently: {kernel_size}) because it might lead to scenarios where the 16-bit count matrix is overflown'

    prediction = input_data[0]
    probability = input_data[1]

    # Build a class mapping, so classes are converted to indexes and vice-versa
    unique_values = set(np.unique(prediction))
    unique_values = sorted(unique_values - set(excluded_values))
    index_value_lut = [(k, v) for k, v in enumerate(unique_values)]

    counts = np.zeros(shape=(*prediction.shape, len(unique_values)), dtype=np.uint16)
    probabilities = np.zeros(shape=(*probability.shape, len(unique_values)), dtype=np.uint16)

    # Iterates for each classes
    for cls_idx, cls_value in index_value_lut:

        # Take the binary mask of the interest class, and multiplies by the probabilities
        class_mask = ((prediction == cls_value) * probability).astype(np.uint16)

        # Sets to 0 the class scores where the threshold is lower
        class_mask[probability <= conf_threshold] = 0

        # Set to 0 the class scores where the label is excluded
        for excluded_value in excluded_values:
            class_mask[prediction == excluded_value] = 0

        # Binary class mask, used to count HOW MANY neighbours pixels are used for this class
        binary_class_mask = (class_mask > 0).astype(np.uint16)

        # Creates the kernel
        kernel = np.ones(shape=(kernel_size, kernel_size), dtype=np.uint16)

        # Counts around the window the sum of probabilities for that given class
        counts[:, :, cls_idx] = convolve2d(class_mask, kernel, mode='same')

        # Counts the number of neighbors pixels that voted for that given class
        class_voters = convolve2d(binary_class_mask, kernel, mode='same')
        class_voters[class_voters == 0] = 1  # Remove the 0 values because might create divide by 0 issues

        probabilities[:, :, cls_idx] = np.divide(counts[:, :, cls_idx], class_voters)

    # Initializes output array
    aggregated_predictions = np.zeros(shape=(counts.shape[0], counts.shape[1]), dtype=np.uint16)
    # Initializes prediction output array
    aggregated_probabilities = np.zeros(shape=(counts.shape[0], counts.shape[1]), dtype=np.uint16)

    if len(unique_values) > 0:
        # Takes the indices that have the biggest scores
        aggregated_predictions_indices = np.argmax(counts, axis=2)

        # Get the new confidence score for the indices
        aggregated_probabilities = np.take_along_axis(
            probabilities,
            aggregated_predictions_indices.reshape(*aggregated_predictions_indices.shape, 1),
            axis=2
        ).squeeze()

        # Check which pixels have a counts value to 0
        no_score_mask = np.sum(counts, axis=2) == 0

        # convert back to values from indices
        for (cls_idx, cls_value) in index_value_lut:
            aggregated_predictions[aggregated_predictions_indices == cls_idx] = cls_value
            aggregated_predictions = aggregated_predictions.astype(np.uint16)

        aggregated_predictions[no_score_mask] = target_excluded_value
        aggregated_probabilities[no_score_mask] = 0

    # Setting excluded values back to their original values
    for excluded_value in excluded_values:
        aggregated_predictions[prediction == excluded_value] = excluded_value
        aggregated_probabilities[prediction == excluded_value] = 0

    return np.array(
        [aggregated_predictions, aggregated_probabilities, *input_data[2:]]
    )


def compute_adaptive_threshold(
    input_data: np.ndarray,
    target_excluded_percentage: int,
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES
) -> int:
    """
    Computes the threshold to have the excluded percentage set as target.
    :param input_data: A numpy array containing in the first dimensions two
        elements: the class codes and the associated predicted
        probability/confidence.
    :param target_excluded_percentage: The percentage of excluded fields to
        target. Threshold will be computed to be as close of this value.
    :param excluded_values: Values in the crop type layer that are excluded and
        do not count in the probability distribution.

    :return: An integer threshold value, between 0 and 100 that yields an
    undecided crop percentage as close as possible from the target.
    """
    croptype = input_data[0]
    probability = input_data[1]

    # Compute pixels that are not excluded
    excluded_mask = np.isin(croptype, excluded_values)

    # Get 1D distribution of max_probability for valid crop pixels
    probability_distribution = np.sort(probability[~excluded_mask])

    target_excluded_percentage_float = float(target_excluded_percentage) / 100.0

    # Binary search on threshold to obtain the nan percentage closest from
    # objective
    percentage_undecided_scores = []

    for threshold in range(0, 101):
        nan_percentage = (probability_distribution < threshold).sum() /\
            len(probability_distribution)
        percentage_undecided_scores.append(nan_percentage)

    percentage_undecided_scores = np.array(percentage_undecided_scores)

    percentage_distance = abs(
        percentage_undecided_scores - target_excluded_percentage_float
    )

    best_threshold = np.argmin(percentage_distance)

    return int(best_threshold)


def set_undecided_by_threshold(
    input_data: np.ndarray,
    threshold_decision: dict,
    threshold: int = 35,
    target_undecided_percentage: int = -1,
    undecided_arable_value: int = DEFAULT_UNDECIDED_ARABLE_VALUE,
    undecided_perennial_value: int = DEFAULT_UNDECIDED_PERENNIAL_VALUE,
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES
) -> np.ndarray:
    """
    Sets the pixels with prediction probabilities under the defined threshold
    value to the target undecided value.

    :param input_data: A 3D numpy array containing in the first axis at least
        the class mask and the probability mask.
    :param threshold: The minimum prediction confidence value under which the
        pixels are sets to the undecided class. Not necessary to specify if
        target_undecided_percentage function is used.
    :param target_undecided_percentage: Defines the maximum percentage of
        undecided pixels. If the given threshold yields a ratio of undecided
        pixels higher than this value, then a threshold is computed to approach
        this limit as much as possible. Set this value to -1 to disable it.
    :param undecided_value_arable: Value to set all undecided arable
        fields to. HRL codes between this variable and the
        parameter `target_excluded_value_perennial` will be
        considered as arable.
    :param undecided_value_perennial: Value to set all undecided perennial
        fields to. All HRL codes above are
        considered as perennial.

    :return: Two numpy arrays. The first containing the new class mask with some
             undecided pixels and the second containing the original prediction
             probability/confidence values.
    """
    prediction = input_data[0]
    probability = input_data[1]

    output_prediction = prediction.copy()

    excluded_mask = np.isin(prediction, excluded_values)

    undecided_mask = probability < threshold
    # compute the value and checks if the target percentage was overeached or
    # not
    undecided_percentage = undecided_mask.sum() \
        / (~excluded_mask).sum()

    logger.info(f'Undecided ratio: {undecided_percentage} with threshold: {threshold}')
    if target_undecided_percentage > -1:
        if undecided_percentage > (target_undecided_percentage / 100.0):
            old_threshold = threshold
            threshold = compute_adaptive_threshold(
                input_data, target_undecided_percentage
            )
            logger.info(f'Threshold: {old_threshold} gives an amount of undecided pixels above the specified limit. Lowering it to: {threshold}')
            undecided_mask = probability < threshold

    # Saving threshold for outer use
    threshold_decision['value'] = threshold

    arable_undecided_mask = undecided_mask & \
        (prediction >= 1000) & \
        (prediction < 2000)
    perennial_undecided_mask = undecided_mask & \
        (prediction >= 2000)

    output_prediction[arable_undecided_mask] = undecided_arable_value
    output_prediction[perennial_undecided_mask] = undecided_perennial_value

    # Keep the excluded values excluded
    output_prediction[excluded_mask] = prediction[excluded_mask]

    return np.array([output_prediction, probability, *input_data[2:]])


def fix_interannual_consistency(
    raw_tiles: np.ndarray,
    postproc_tiles: np.ndarray,
    class_ids: List[int] = list(DEFAULT_HRL_NAMES_TABLE.keys()),
    excluded_values: List[int] = DEFAULT_EXCLUDED_VALUES + [GRASS_AND_FODDER_CODE]
) -> np.ndarray:
    """ Performs a interannnual consistency correction, a few cases are cover:
        - When a sequence of croptypes predictions predict only perennial, except
        for one observations (that is not in the edges) then corrects that
        sequence to the perennial class that has the maximum accumulated
        probability over the sequence.
        - When a full sequence of perennials does not have the same crop type,
        then sets that sequence with the perennial class that has the maximum
        accumulated probability over the sequence.
        - When there is a pixel of perennial surrounded by two arables, which
        most often corresponds to a fallow land. Those pixels are set to NoCrop.

        Parameters
        ----------
        raw_tles: np.ndarray
            A 4D array containing in the first axis the year
            dimension, then the channels of the raw input tile and then the x
            and y axis.
        postproc_tiles: np.ndarray
            An array corresponding the same amount of
            dimension as the raw tiles, containing for each year the
            postprocessed output class and the associated max probability of the
            tile.
        class_ids: List[int]
            List of the class values valid for those tiles, need
            to be the same size as the number of per-class probability in the
            raw tiles, and needs to be in the same order.
        excluded_values: List[int]
            List of values of pixels to not consider here.

        Returns
        -------
        output: np.ndarray
            A ndarray containing for each year the final fixed croptype and
            associated max_probability.
    """ 
    
    class_ids = np.array(class_ids)

    n_years = raw_tiles.shape[0]
    raw_class_probs = raw_tiles[:, 2:, :, :]
    postproc_croptype = postproc_tiles[:, 0, :, :]


    # Compute an excluded mask
    excluded_mask = np.zeros(raw_class_probs.shape[2:], dtype='bool')
    for value in excluded_values:
        excluded_mask = np.bitwise_or(
            excluded_mask, (postproc_croptype == value).any(axis=0)
        )

    # Computes the valid mask a the opposite of the excluded mask
    valid_mask = ~excluded_mask

    # Gets the count of anual crop occurencies
    arable_mask = np.bitwise_and(
        postproc_croptype < 2000, valid_mask
    )

    perennial_mask = np.bitwise_and(
        postproc_croptype >= 2000, valid_mask
    )
    perennial_count_mask = perennial_mask.astype('uint16').sum(axis=0)

    output = postproc_tiles.copy()

    # Changes to (bands, year, x, y)
    output = np.transpose(output, axes=(1, 0, 2, 3))


    # FIRST CASE ---------------------------------------------------------------

    # Must be a single missng perennial observation but not at the beginning or
    # end of the sequence, this mask will be added to the different_values_mask
    # in the second phase

    single_perennial_hole = np.bitwise_and(
        perennial_count_mask == 4,
        perennial_mask[0, :, :],
    )
    single_perennial_hole = np.bitwise_and(
        single_perennial_hole,
        perennial_mask[-1, :, :]
    )

    # SECOND CASE --------------------------------------------------------------

    # For all the pixels that have consitent perennial classes, make sure that
    # the predicted perennial classes are the same.

    all_perennial = np.all(
        perennial_mask, axis=0
    )
    different_values_mask = ~np.all(
        postproc_croptype == postproc_croptype[0, ...],
        axis=0
    )
    different_values_mask = np.bitwise_and(
        different_values_mask, all_perennial
    )

    different_values_mask = np.bitwise_or(
        different_values_mask, single_perennial_hole
    )

    # For every pixel with a different sequence, sum all probabilities of all 
    # perennials and sets the final perennial class to the winning probability 
    # class.
    summed_probs_perennial = raw_class_probs[:, 14:, :, :].sum(axis=0)
    output[0, :, :, :][:, different_values_mask] = class_ids[
        np.argmax(summed_probs_perennial[:, different_values_mask], axis=0) + 14
    ]
    output[1, :, :, :][:, different_values_mask] = (
        (summed_probs_perennial[:, different_values_mask].max(axis=0) / n_years)
        + 0.5
    ).astype('uint8')

    # THIRD CASE ---------------------------------------------------------------
    
    # In our 3D binary mask of arable/non arable let's see which years a non
    # arable pixel is surrounded by arable

    # For each year counts within the previous, current and next year
    # the number of years where a pixel was detected as arable
    for year_index in range(0, n_years):
        if year_index == 0:
            prev_arable = np.ones(arable_mask.shape[1:], dtype='bool')
        else:
            prev_arable = arable_mask[year_index - 1]

        current_perennial = perennial_mask[year_index]

        if year_index == n_years - 1:
            next_arable = np.ones(arable_mask.shape[1:], dtype='bool')
        else:
            next_arable = arable_mask[year_index + 1]

        fallow_pixels_mask = prev_arable & current_perennial & next_arable
        output[0, year_index, fallow_pixels_mask] = DEFAULT_NO_CROP
        output[1, year_index, fallow_pixels_mask] = DEFAULT_NO_CROP_PROBABILITY

    # Changes back to orignal (years, bands, x, y)
    output = np.transpose(output, axes=(1, 0, 2, 3))

    return output
